import io
import pickle
from typing import Literal

import implicit
import numpy as np
import torch
import wandb
from matplotlib import pyplot as plt
from sklearn.decomposition import TruncatedSVD

from my_tools.utils import build_class

from ..interface import (
    RecommenderModuleBase,
    FitExplicitInterfaceMixin,
    UnpopularRecommenderMixin,
)
from ..utils import wandb_plt_figure, Timer


class RandomRecommender(RecommenderModuleBase, FitExplicitInterfaceMixin):
    def fit(self):
        pass

    def forward(self, user_ids, item_ids):
        ratings = torch.randn(len(user_ids), len(item_ids))
        return ratings


class PopularRecommender(RecommenderModuleBase, FitExplicitInterfaceMixin):
    items_count: torch.Tensor

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.register_buffer(name="items_count", tensor=torch.zeros(self.n_items))

    def fit(self):
        implicit_feedback = self.to_scipy_coo(self.explicit) > 0
        self.items_count = torch.from_numpy(implicit_feedback.sum(axis=0).A.squeeze(0))

    def forward(self, user_ids, item_ids):
        ratings = self.items_count[item_ids].repeat(len(user_ids), 1)
        return ratings.to(torch.float32)


class SVDRecommender(RecommenderModuleBase, FitExplicitInterfaceMixin):
    def __init__(self, n_components=10, **kwargs):
        super().__init__(**kwargs)
        self.model = TruncatedSVD(n_components=n_components)

    def fit(self):
        self.model.fit(self.to_scipy_coo(self.explicit))
        self.plot_explained_variance()

    def plot_explained_variance(self):
        if wandb.run is None:
            return
        wandb.log(
            dict(explained_variance_ratio=self.model.explained_variance_ratio_.sum())
        )
        with wandb_plt_figure(
            title="Explained variance depending on number of components"
        ):
            plt.xlabel("Number of components")
            plt.ylabel("Explained cumulative variance ratio")
            plt.plot(np.cumsum(self.model.explained_variance_ratio_))

    @Timer()
    def online_ratings(self, users_explicit):
        users_explicit = self.to_scipy_coo(users_explicit)
        embedding = self.model.transform(users_explicit)
        ratings = self.model.inverse_transform(embedding)
        return torch.from_numpy(ratings).to(torch.float32)

    def get_extra_state(self):
        pickled_bytes = pickle.dumps(self.model)
        return pickled_bytes

    def set_extra_state(self, pickled_bytes):
        self.model = pickle.load(io.BytesIO(pickled_bytes))


class ImplicitRecommenderBase(RecommenderModuleBase, FitExplicitInterfaceMixin):
    def __init__(self, *, implicit_model, implicit_kwargs=None, **kwargs):
        super().__init__(**kwargs)
        self.model = build_class(
            module_candidates=[
                implicit.nearest_neighbours,
                implicit.als,
                implicit.bpr,
                implicit.lmf,
            ],
            class_name=implicit_model,
            **(implicit_kwargs or {}),
        )

    def fit(self):
        self.explicit = self.explicit.to(torch.float32)
        self.model.fit(self.to_scipy_coo(self.explicit))

    def forward(self, user_ids, item_ids):
        explicit_feedback = self.to_scipy_coo(self.explicit).tocsr()
        recommended_item_ids, item_ratings = self.model.content(
            userid=user_ids.numpy(),
            user_items=explicit_feedback[user_ids],
            N=self.n_items,
            filter_already_liked_items=False,
        )
        ratings = np.full(
            shape=(len(user_ids), self.n_items + 1),
            fill_value=np.finfo(np.float32).min,
        )
        np.put_along_axis(
            arr=ratings, indices=recommended_item_ids, values=item_ratings, axis=1
        )
        ratings = torch.from_numpy(ratings[:, :-1])
        return ratings[:, item_ids]

    @torch.inference_mode()
    def recommend(self, user_ids, n_recommendations=None):
        item_ids, item_ratings = self.model.recommend(
            userid=user_ids.numpy(),
            user_items=self.to_scipy_coo(self.explicit).tocsr()[user_ids],
            N=n_recommendations or self.n_items,
            filter_already_liked_items=True,
        )
        item_ids = torch.from_numpy(item_ids).to(torch.int64)
        self.check_invalid_recommendations(recommendations=item_ids, warn=False)
        return item_ids

    def get_extra_state(self):
        bytesio = io.BytesIO()
        self.model.save(bytesio)
        return bytesio

    def set_extra_state(self, bytesio):
        bytesio.seek(0)
        self.model = self.model.load(bytesio)


class ImplicitNearestNeighborsRecommender(ImplicitRecommenderBase):
    def __init__(
        self,
        implicit_model: Literal[
            "BM25Recommender", "CosineRecommender", "TFIDFRecommender"
        ] = "BM25Recommender",
        num_neighbors=20,
        num_threads=0,
        **kwargs,
    ):
        super().__init__(
            implicit_model=implicit_model,
            implicit_kwargs=dict(K=num_neighbors, num_threads=num_threads),
            **kwargs,
        )

    def online_recommend(
        self, users_explicit, n_recommendations=None, **kwargs
    ) -> torch.IntTensor:
        users_explicit = self.to_torch_coo(users_explicit)
        item_ids, item_ratings = self.model.recommend(
            userid=np.arange(users_explicit.shape[0]),
            user_items=self.to_scipy_coo(users_explicit.to(torch.float32)).tocsr(),
            N=n_recommendations or self.n_items,
            **kwargs,
        )
        item_ids = torch.from_numpy(item_ids).to(torch.int64)
        self.check_invalid_recommendations(recommendations=item_ids, warn=False)
        return item_ids


class ImplicitMatrixFactorizationRecommender(ImplicitRecommenderBase):
    def __init__(
        self,
        implicit_model: Literal[
            "AlternatingLeastSquares",
            "LogisticMatrixFactorization",
            "BayesianPersonalizedRanking",
        ] = "AlternatingLeastSquares",
        factors=100,
        learning_rate=1e-2,
        regularization=1e-2,
        num_threads=0,
        use_gpu=True,
        implicit_kwargs=None,
        **kwargs,
    ):
        implicit_kwargs = implicit_kwargs or {}
        implicit_kwargs.update(
            factors=factors,
            learning_rate=learning_rate,
            regularization=regularization,
            num_threads=num_threads,
            use_gpu=use_gpu,
        )
        if implicit_model == "AlternatingLeastSquares":
            implicit_kwargs.pop("learning_rate")
        else:
            implicit_kwargs["use_gpu"] = False
        super().__init__(
            implicit_model=implicit_model, implicit_kwargs=implicit_kwargs, **kwargs
        )


class UnpopularSVDRecommender(SVDRecommender, UnpopularRecommenderMixin):
    def __init__(self, *args, unpopularity_coef=1e-3, **kwargs):
        super().__init__(*args, **kwargs)
        self.init_unpopular_recommender_mixin(unpopularity_coef=unpopularity_coef)

    def fit(self):
        self.fit_unpopular_recommender_mixin()
        return super().fit()

    def online_ratings(self, users_explicit, users_activity=None):
        ratings = super().online_ratings(users_explicit=users_explicit)
        if users_activity is None:
            users_activity = torch.from_numpy(
                (self.to_scipy_coo(users_explicit) > 0).mean(1).A.squeeze(1)
            )
        ratings += self.additive_rating_offset(users_activity=users_activity)
        return ratings



import abc
import io

import numba
import numpy as np
import pandas as pd
import torch
import wandb
from scipy.sparse import spmatrix


from pandas.io.formats.style import Styler
from tqdm.auto import tqdm
from typing_extensions import Literal

from my_tools.models import WandbLoggerMixin

from ..interface import (
    RecommenderModuleBase,
    FitExplicitInterfaceMixin,
    ExplanationMixin,
)
from ..utils import build_weight, build_bias


class ALSInterface:
    """Skeleton to expose main als logic."""

    @abc.abstractmethod
    def fit(self):
        """Where "alternating" in als comes from."""
        for epoch in range(10):
            self.least_squares_optimization_with_fixed_factors(fixed="items")
            self.least_squares_optimization_with_fixed_factors(fixed="users")

    def least_squares_optimization_with_fixed_factors(
        self, fixed: Literal["users", "items"]
    ):
        """Decoupler function to enable dispatcher independence on what features are currently fixed"""
        kwargs = self.preprocess_optimization_args(fixed=fixed)
        self.analytic_optimum_dispatcher(**kwargs)

    @abc.abstractmethod
    def preprocess_optimization_args(self, *, fixed: Literal["users", "items"]) -> dict:
        """
        Prepare arguments for dispatcher to facilitate its
        independence on what features are currently fixed.
        """

    @abc.abstractmethod
    def analytic_optimum_dispatcher(self, **kwargs) -> None:
        """
        Main optimization logic to calculate analytic
        MSE minimum with one of users or items features fixed.
        This method is agnostic about what features are fixed.
        Where "least squares" in als comes from.
        """


class ALS(
    ALSInterface,
    RecommenderModuleBase,
    FitExplicitInterfaceMixin,
    WandbLoggerMixin,
    ExplanationMixin,
):
    def __init__(
        self,
        epochs=10,
        latent_dimension_size=10,
        regularization_lambda=100,
        confidence_alpha=10,
        lambda_decay=0.75,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.epochs = epochs
        self.regularization_lambda = regularization_lambda
        self.confidence_alpha = confidence_alpha
        self.lambda_decay = lambda_decay

        self.user_factors = build_weight(self.n_users, latent_dimension_size)
        self.user_factors.requires_grad = False
        self.item_factors = build_weight(self.n_items, latent_dimension_size)
        self.item_factors.requires_grad = False

        self.confidence_x_preference = None
        self.confidence_minus_1 = None
        self.preference = None

    def init_preference_confidence(self, explicit_feedback: spmatrix):
        implicit_feedback = (explicit_feedback > 0).astype(np.float32)
        self.preference = implicit_feedback
        self.confidence_minus_1 = implicit_feedback * self.confidence_alpha
        self.confidence_minus_1.eliminate_zeros()
        self.confidence_x_preference = (
            self.confidence_minus_1.multiply(self.preference) + self.preference
        )

    def fit(self):
        self.init_preference_confidence(self.to_scipy_coo(self.explicit))
        for _ in tqdm(range(self.epochs), "Alternating"):
            self.on_train_epoch_start()
            self.least_squares_optimization_with_fixed_factors(fixed="items")
            self.least_squares_optimization_with_fixed_factors(fixed="users")
            self.regularization_lambda *= self.lambda_decay

    def on_train_epoch_start(self):
        self.log(dict(regularization_lambda=self.regularization_lambda))

    def least_squares_optimization_with_fixed_factors(
        self, fixed: Literal["users", "items"]
    ):
        kwargs = self.preprocess_optimization_args(fixed=fixed)
        self.analytic_optimum_dispatcher(**kwargs)

    def preprocess_optimization_args(self, *, fixed):
        if fixed == "items":
            X = self.user_factors.numpy()
            Y = self.item_factors.numpy()
            sparse_iterator = self.sparse_iterator()
        elif fixed == "users":
            X = self.item_factors.numpy()
            Y = self.user_factors.numpy()
            sparse_iterator = self.sparse_iterator(transpose=True)
        else:
            raise ValueError
        return dict(X=X, Y=Y, sparse_iterator=sparse_iterator, fixed=fixed)

    def analytic_optimum_dispatcher(self, X, Y, sparse_iterator, fixed):
        """This implementation and the next helper method are left here mainly for
        completeness's sake, for more efficient and less convoluted implementation see ALSJIT."""
        lambda_I = self.regularization_lambda * np.eye(Y.shape[1])
        YtY_plus_lambdaI = Y.T @ Y + lambda_I

        for row_id, col_indices, cm1, cp in sparse_iterator:
            if col_indices.size == 0:
                X[row_id] = 0
                continue

            y = Y[col_indices]
            YtCY_plus_lambdaI = (y.T * cm1) @ y + YtY_plus_lambdaI
            X[row_id] = np.linalg.inv(YtCY_plus_lambdaI) @ (y.T @ cp)

    def sparse_iterator(self, transpose=False):
        confidence_minus_1 = self.confidence_minus_1
        confidence_x_preference = self.confidence_x_preference

        if transpose:
            confidence_minus_1 = confidence_minus_1.tocsc()
            confidence_x_preference = confidence_x_preference.tocsc()

        cm1_data = confidence_minus_1.data
        cp_data = confidence_x_preference.data
        indices = confidence_minus_1.indices
        indptr = confidence_minus_1.indptr
        for ptr_id, (ind_begin, ind_end) in enumerate(zip(indptr, indptr[1:])):
            ind_slice = slice(ind_begin, ind_end)
            yield ptr_id, indices[ind_slice], cm1_data[ind_slice], cp_data[ind_slice]

    def forward(self, user_ids, item_ids):
        ratings = self.user_factors[user_ids] @ self.item_factors[item_ids].T
        return ratings

    def explain_recommendations_for_user(
        self,
        user_id=None,
        user_explicit=None,
        n_recommendations=10,
        log=False,
        logging_prefix="",
    ) -> Styler:
        if user_id is None:
            raise NotImplementedError(
                "Online explanations are not implemented for als."
            )
        recommendations = self.recommend(
            user_ids=torch.IntTensor([user_id]), n_recommendations=n_recommendations
        ).numpy()
        user_ratings = self(user_ids=torch.tensor([user_id]), item_ids=recommendations)[
            0
        ].numpy()
        YtY_plus_lambdaI = item_factors_regularization_term = (
            self.item_factors.T @ self.item_factors
            + self.regularization_lambda * np.eye(self.item_factors.shape[1])
        ).numpy()
        relative_row_ids, liked_item_ids = self.preference[user_id].nonzero()
        if liked_item_ids.size == 0:
            raise ValueError(
                "No personal recommendations generated, probably a cold user."
            )

        confidence_minus_1 = (
            self.confidence_minus_1[user_id, liked_item_ids].toarray().squeeze()
        )
        Y = liked_items_factors = self.item_factors[liked_item_ids].numpy()
        Y_recommended = recommended_items_factors = self.item_factors[
            recommendations
        ].numpy()
        user_latent_weight = YtCY = np.linalg.inv(
            (Y.T * confidence_minus_1) @ Y + YtY_plus_lambdaI
        )
        similarity = Y_recommended @ user_latent_weight @ Y.T

        confidence = confidence_minus_1 + 1

        explanations = {
            "ratings": pd.Series(
                user_ratings,
                recommendations,
                name=f"predicted relevance",
            ).rename_axis(index="recommended items for user"),
            "similarity": pd.DataFrame(
                similarity, index=recommendations, columns=liked_item_ids
            ).rename_axis(
                index=f"recommended items for user",
                columns=f"liked items by user",
            ),
            "confidence": pd.Series(
                confidence,
                liked_item_ids,
                name=f"estimated confidence that user likes items",
            ).rename_axis(index="liked items by user"),
        }

        dataframe = pd.concat(
            [explanations["similarity"], explanations["ratings"]], axis="columns"
        )
        dataframe = pd.concat(
            [dataframe.T, explanations["confidence"]], axis="columns"
        ).T.rename_axis(index="recommended items", columns="liked items")

        style = dataframe.style
        style.set_caption(
            f"Recommended items for user {user_id} based on personal similarity measure "
            "between items and estimated confidence in these similarities."
        ).set_table_styles(
            [
                dict(
                    selector="caption",
                    props=[
                        ("text-align", "center"),
                        ("font-size", "125%"),
                        ("color", "black"),
                    ],
                )
            ]
        )
        common_gradient_kwargs = dict(low=0.5, high=0.5)
        style = style.background_gradient(
            subset=(dataframe.index[-1], dataframe.columns[:-1]),
            axis="columns",
            **common_gradient_kwargs,
            cmap="YlOrRd",
        )
        style = style.background_gradient(
            subset=(dataframe.index[:-1], dataframe.columns[:-1]),
            **common_gradient_kwargs,
            cmap="coolwarm",
        )
        style = style.background_gradient(
            subset=(dataframe.index[:-1], dataframe.columns[-1]),
            **common_gradient_kwargs,
            cmap="YlOrRd",
        )
        if log:
            textio = io.TextIOWrapper(io.BytesIO())
            style.to_html(textio)
            textio.seek(0)
            wandb.log(
                {
                    logging_prefix
                    + " Recommendation explaining dataframe": wandb.Html(textio)
                }
            )
        return style


class ALSjit(ALS):
    def __init__(self, *args, num_threads=8, **kwargs):
        super().__init__(*args, **kwargs)
        numba.config.NUMBA_NUM_THREADS = 32
        numba.config.THREADING_LAYER = "threadsafe"
        numba.set_num_threads(num_threads)

    def preprocess_optimization_args(self, **kwargs):
        kwargs = super().preprocess_optimization_args(**kwargs)

        if kwargs["fixed"] == "items":
            confidence_minus_1 = self.confidence_minus_1
            confidence_x_preference = self.confidence_x_preference
        elif kwargs["fixed"] == "users":
            confidence_minus_1 = self.confidence_minus_1.tocsc()
            confidence_x_preference = self.confidence_x_preference.tocsc()
        else:
            raise ValueError("Unknown fixed value")

        cm1_data = confidence_minus_1.data.astype(np.float32)
        cp_data = confidence_x_preference.data.astype(np.float32)
        indices = confidence_minus_1.indices
        indptr = confidence_minus_1.indptr

        X = kwargs["X"]
        Y = kwargs["Y"]
        lambda_I = self.regularization_lambda * np.eye(Y.shape[1], dtype=np.float32)
        YtY_plus_lambdaI = Y.T @ Y + lambda_I

        return dict(
            X=X,
            Y=Y,
            cm1_data=cm1_data,
            cp_data=cp_data,
            indices=indices,
            indptr=indptr,
            YtY_plus_lambdaI=YtY_plus_lambdaI,
        )

    @staticmethod
    @numba.njit(parallel=True)
    def analytic_optimum_dispatcher(
        X, Y, cm1_data, cp_data, indices, indptr, YtY_plus_lambdaI
    ):
        for row_id in numba.prange(X.shape[0]):
            ind_slice = slice(indptr[row_id], indptr[row_id + 1])
            if ind_slice.start == ind_slice.stop:
                X[row_id] = 0
                continue

            cm1 = cm1_data[ind_slice]
            cp = cp_data[ind_slice]
            col_indices = indices[ind_slice]
            y = Y[col_indices]

            YtCY_plus_lambdaI = (y.T * cm1) @ y + YtY_plus_lambdaI
            X[row_id] = np.linalg.inv(YtCY_plus_lambdaI).astype(np.float32) @ (y.T @ cp)


class ALSjitBiased(ALSjit):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.bias = build_bias(1)
        self.bias.requires_grad = False

        self.user_factors = build_weight(
            self.user_factors.shape[0], self.user_factors.shape[1] + 2
        )
        self.user_factors.requires_grad = False
        self.user_factors[:, 0] = 1

        self.item_factors = build_weight(
            self.item_factors.shape[0], self.item_factors.shape[1] + 2
        )
        self.item_factors.requires_grad = False
        self.item_factors[:, 1] = 1

    def on_train_epoch_start(self):
        super().on_train_epoch_start()
        self.log(
            dict(
                bias=self.bias.item(),
                mean_user_bias=self.user_factors[:, 1].mean(),
                mean_item_bias=self.item_factors[:, 0].mean(),
            )
        )

    def preprocess_optimization_args(self, **kwargs):
        fixed = kwargs["fixed"]
        kwargs = super().preprocess_optimization_args(**kwargs)

        if fixed == "items":
            X_constant_latent_index = 0
        elif fixed == "users":
            X_constant_latent_index = 1
        else:
            raise ValueError("Unknown fixed value")

        YtY_plus_lambdaI = kwargs["YtY_plus_lambdaI"]
        YtY_plus_lambdaI[
            X_constant_latent_index, X_constant_latent_index
        ] -= self.regularization_lambda

        XY_sum = (
            (self.user_factors.sum(axis=0) * self.item_factors.sum(axis=0)).sum().item()
        )

        self.bias[:] = (
            self.confidence_x_preference.sum()
            - XY_sum
            - (
                self.user_factors.T.numpy()
                @ self.confidence_minus_1
                @ self.item_factors.numpy()
            ).sum()
        ) / (self.confidence_minus_1.sum() + np.prod(self.confidence_minus_1.shape))

        return {
            **{
                k: kwargs[k]
                for k in [
                    "X",
                    "Y",
                    "cm1_data",
                    "cp_data",
                    "indices",
                    "indptr",
                ]
            },
            **{
                "YtY_plus_lambdaI": YtY_plus_lambdaI,
                "bias": self.bias.numpy(),
                "X_constant_latent_index": X_constant_latent_index,
            },
        }

    @staticmethod
    @numba.njit(parallel=True)
    def analytic_optimum_dispatcher(
        X,
        Y,
        cm1_data,
        cp_data,
        indices,
        indptr,
        YtY_plus_lambdaI,
        bias,
        X_constant_latent_index,
    ):
        for row_id in numba.prange(X.shape[0]):
            ind_slice = slice(indptr[row_id], indptr[row_id + 1])
            if ind_slice.start == ind_slice.stop:
                X[row_id] = 0
                continue

            cm1 = cm1_data[ind_slice]
            cp = cp_data[ind_slice]
            col_indices = indices[ind_slice]
            y = Y[col_indices]

            YtCY_plus_lambdaI = (y.T * cm1) @ y + YtY_plus_lambdaI
            YtCP_minus_biasYtC = y.T @ (cp - bias[0] * cm1 - bias[0])
            X[row_id] = np.linalg.inv(YtCY_plus_lambdaI).astype(np.float32) @ (
                YtCP_minus_biasYtC
            )

        X[:, X_constant_latent_index] = 1

    def explain_recommendations_for_user(self, *args, **kwargs):
        raise NotImplementedError("Need to add bias to existing implementation")



import einops
import numpy as np
import pandas as pd
import torch

from numba import njit
from scipy.sparse import csr_matrix


def relevant_pairs_to_frame(relevant_pairs):
    relevant_pairs = pd.DataFrame(relevant_pairs)
    relevant_pairs.columns = ["user_id", "item_id"]
    return relevant_pairs


@njit
def mark_duplicate_recommended_items_as_invalid(recommendations, invalid_mark=-1):
    """Mark all but first duplicated items as invalid for each user.
    :param recommendations: array of shape [n_users, n_items] with recommended item_ids
    :param invalid_mark: integer to replace duplicates with
    :return: recommendations array with duplicated values in each row marked as invalid,
    keeping the first occurrence.
    """
    for user, recommended_items in enumerate(recommendations):
        occurred_item_ids = []
        for item_rank, item in enumerate(recommended_items):
            if item in occurred_item_ids:
                recommendations[user, item_rank] = invalid_mark
            else:
                occurred_item_ids.append(item)
    return recommendations


def binary_relevance(explicit_feedback, recommendee_user_ids, recommendations):
    """
    Determine whether recommended items are relevant to users based on provided relevant pairs.
    :param explicit_feedback: csr_matrix with explicit feedback
    :param recommendee_user_ids: numpy array of shape [n_users] indicating users for whom the
    recommendations were generated.
    :param recommendations: numpy array of shape [n_users, n_items] with recommended item ids,
    with each row corresponding to user_id from recommendee_user_ids, with items sorted in
    decreasing predicted relevance.
    :return: numpy bool array of shape [n_users, n_items] indicating whether the recommended
    [user_id, item_id] pair is in relevant_pairs.
    """
    explicit = explicit_feedback[recommendee_user_ids]
    relevance = njit_binary_relevance(
        explicit.indptr, explicit.indices, recommendations
    )
    return relevance


@njit
def njit_binary_relevance(indptr, indices, recommendations):
    relevance = np.zeros_like(recommendations)
    for i, (begin, end, item_ids) in enumerate(
        zip(indptr, indptr[1:], recommendations)
    ):
        for j, item_id in enumerate(item_ids):
            relevance[i, j] = item_id in indices[begin:end]
    return relevance


def hitrate(relevance):
    """
    Returns proportion of users who were hit (were recommended at least one relevant item).
    This metric is pretty basic and represents an upper bound on other metrics
    and the model performance in general.

    :param relevance: boolean array of shape [n_users, n_items], representing
    relevant recommendations per user.
    """
    hits_per_user = einops.reduce(relevance, "user relevance -> user", np.any)
    return np.mean(hits_per_user)


def accuracy(relevance):
    """Average proportion of relevant recommendations."""
    return np.mean(relevance)


def mean_reciprocal_rank(relevance):
    """
    Returns average reciprocal rank, i.e. 1 / position, of first relevant item per user.
    Represents how far users have to explore their recommendations before encountering
    something relevant.
    """
    reciprocal_ranks = 1 / np.arange(1, relevance.shape[1] + 1)
    relevant_ranks = np.einsum("i j, j -> i j", relevance, reciprocal_ranks)
    first_relevant_rank = einops.reduce(relevant_ranks, "user rank -> user", np.max)
    return np.mean(first_relevant_rank)


def number_of_all_relevant_items_per_user(relevant_pairs, recommendee_user_ids):
    return (
        relevant_pairs_to_frame(relevant_pairs)
        .groupby("user_id")
        .size()
        .loc[recommendee_user_ids]
        .fillna(0)
        .astype(np.int32)
        .to_numpy()
    )


def recall(relevance, n_relevant_items_per_user):
    """
    Proportion of retrieved relevant items among all
    known to be relevant for each user.
    Very sensitive to size of recommendations.
    """
    return np.mean(
        einops.reduce(relevance, "user item -> user", "sum")
        / np.where(n_relevant_items_per_user == 0, 1, n_relevant_items_per_user)
    )


def mean_average_precision(relevance, n_relevant_items_per_user):
    """
    This metric represents average proportion of relevant items at every possible cutoff.
    While metrics generally try to keep their image equal to the [0, 1] interval, value of 1 is
    unreachable if user has no relevant items. This metric suffers from this downside the
    most, so it needs to know maximum possible number of relevant items per user to
    normalize its image to the [0, 1] interval.
    """

    reciprocal_ranks = 1 / np.arange(1, relevance.shape[1] + 1)
    n_relevant_items_at_k = np.cumsum(relevance, axis=1)
    precision_at_k = np.einsum("i j, j -> i j", n_relevant_items_at_k, reciprocal_ranks)
    average_precision = np.einsum(
        "i j, i j, i -> i",
        precision_at_k,
        relevance,
        1 / np.maximum(n_relevant_items_per_user, 1e-8),
    )
    return np.mean(average_precision)


def coverage(recommendations, all_possible_item_ids):
    """
    Coverage represents the proportion of items that ever get recommended.
    This metric is very sensitive to the structure of recommendations, so it is
    more of a debug kind of metric and useful only when comparing models
    with same queried users and same number of recommendations.

    :param recommendations: array of shape [n_users, n_items] with recommended item ids
    :param all_possible_item_ids: array or set with all item_ids from training dataset
    """
    all_items = set(all_possible_item_ids)
    return len(set(recommendations.flat) & all_items) / len(all_items)


def normalized_items_self_information(explicit_feedback):
    """
    Computes self information of each item, representing how much information
    an item occurrence brings. The higher the information, the more niche the
    item is.
    :return: pandas Series with item ids in index and self information in values
    """
    explicit_feedback = explicit_feedback.tocoo()
    relevant_pairs = np.stack([explicit_feedback.row, explicit_feedback.col]).T
    relevant_pairs = relevant_pairs_to_frame(relevant_pairs)
    n_item_interactions = relevant_pairs.groupby("item_id").size()
    information = np.log2(len(relevant_pairs)) - np.log2(n_item_interactions)
    assert (
        information.notna().all()
    ), f"Debug: Self information shouldn't be na, {information}"
    return information / np.log2(len(relevant_pairs))


@njit
def jit_surpisal(item_ids, self_information_per_item, recommendations):
    self_information = {i: j for i, j in zip(item_ids, self_information_per_item)}
    recommendations_information = np.zeros_like(recommendations, dtype=np.float32)
    for i, row in enumerate(recommendations):
        for j, item_id in enumerate(row):
            if item_id in self_information:
                recommendations_information[i, j] = self_information[item_id]
    return np.mean(recommendations_information)


def surprisal(recommendations, explicit_feedback=None, items_information=None):
    """
    Computes average normalized self information for each recommended item.
    The higher the surprisal, the more specific recommendations the model gives.

    :param recommendations: recommended item ids per user
    :param explicit_feedback: explicit feedback csr_matrix
    :param items_information: precomputed items self information
    """
    if items_information is None:
        if explicit_feedback is None:
            raise ValueError(
                "Either explicit_feedback or items_information must be passed."
            )
        items_information = normalized_items_self_information(explicit_feedback)
    return jit_surpisal(
        item_ids=items_information.index.to_numpy(),
        self_information_per_item=items_information.values,
        recommendations=recommendations,
    )


def discounted_cumulative_gain(relevance):
    """This DCG implementation works with float relevance as well"""
    relevance = 2**relevance - 1
    discount = 1 / np.log2(np.arange(2, relevance.shape[1] + 2))
    return np.einsum("i j, j -> i", relevance, discount)


def normalized_discounted_cumulative_gain(relevance, n_relevant_items_per_user):
    """General front-heavy measure of ranking effectiveness."""
    dcg = discounted_cumulative_gain(relevance)
    ideal_relevance = np.arange(relevance.shape[1]) < n_relevant_items_per_user[:, None]
    ideal_dcg = discounted_cumulative_gain(ideal_relevance)
    return np.mean(dcg / (ideal_dcg + 1e-8))


class RecommendingMetrics:
    def __init__(self, explicit_feedback: csr_matrix):
        self.explicit_feedback = explicit_feedback
        self.items_information = normalized_items_self_information(explicit_feedback)
        self.all_items = self.items_information.index.to_numpy()
        self.unique_recommended_items = np.empty(0, dtype=np.int32)
        self.n_relevant_items_per_user = (explicit_feedback > 0).sum(axis=1).A1
        self.k = None

    @staticmethod
    def atk_suffix(dictionary, k):
        return {key + "@" + str(k): v for key, v in dictionary.items()}

    def batch_metrics(
        self, user_ids: torch.IntTensor, recommendations: torch.IntTensor
    ):
        user_ids = user_ids.to("cpu", torch.int32).numpy()
        recommendations = recommendations.to("cpu", torch.int32).numpy()
        relevance = binary_relevance(self.explicit_feedback, user_ids, recommendations)
        n_relevant_items_per_user = self.n_relevant_items_per_user[user_ids]
        self.update_coverage(recommendations)
        metrics = dict(
            hitrate=hitrate(
                relevance=relevance,
            ),
            accuracy=accuracy(
                relevance=relevance,
            ),
            recall=recall(
                relevance=relevance,
                n_relevant_items_per_user=n_relevant_items_per_user,
            ),
            ndcg=normalized_discounted_cumulative_gain(
                relevance=relevance,
                n_relevant_items_per_user=n_relevant_items_per_user,
            ),
            map=mean_average_precision(
                relevance=relevance,
                n_relevant_items_per_user=n_relevant_items_per_user,
            ),
            mrr=mean_reciprocal_rank(
                relevance=relevance,
            ),
            surprisal=surprisal(
                recommendations=recommendations,
                items_information=self.items_information,
            ),
        )
        self.k = recommendations.shape[1]
        return self.atk_suffix(metrics, k=self.k)

    def update_coverage(self, recommendations):
        self.unique_recommended_items = np.union1d(
            self.unique_recommended_items, recommendations
        )

    def finalize_coverage(self):
        coverage = len(
            np.intersect1d(self.unique_recommended_items, self.all_items)
        ) / len(self.all_items)
        coverage_atk = self.atk_suffix(dict(coverage=coverage), k=self.k)
        self.unique_recommended_items = np.empty(0, dtype=np.int32)
        self.k = None
        return coverage_atk

    def metrics(self, user_ids: torch.IntTensor, recommendations: torch.IntTensor):
        metrics = self.batch_metrics(user_ids, recommendations)
        self.update_coverage(recommendations)
        coverage = self.finalize_coverage()
        metrics.update(coverage)
        return metrics



import abc
import argparse
import functools
import os
import string
from collections import OrderedDict
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import torch
import wandb

from scipy.sparse import coo_matrix

from ..utils import profile, prepare_artifact

if TYPE_CHECKING:
    from my_tools.utils import SparseTensor
    from .lit import LitRecommenderBase


def csv_imdb_ratings_to_dataframe(
    path_to_imdb_ratings_csv: str, ratings_scale_max_to_convert_to: float = 5
) -> pd.DataFrame:
    dataframe = pd.read_csv(path_to_imdb_ratings_csv)
    dataframe = dataframe.rename(lambda c: c.lower().replace(" ", "_"), axis="columns")
    dataframe = dataframe.rename({"const": "imdb_id"}, axis="columns")
    dataframe["imdb_id"] = dataframe["imdb_id"].str.removeprefix("tt").astype(int)
    dataframe["your_rating"] *= ratings_scale_max_to_convert_to / 10
    assert list(dataframe) == [
        "imdb_id",
        "your_rating",
        "date_rated",
        "title",
        "url",
        "title_type",
        "imdb_rating",
        "runtime_(mins)",
        "year",
        "genres",
        "num_votes",
        "release_date",
        "directors",
    ]
    return dataframe


class RecommendingDatasetInterface(abc.ABC):
    @property
    @abc.abstractmethod
    @functools.lru_cache
    def unique_dataset_user_ids(self) -> np.array:
        """Cached unique sorted user ids that the model can encounter during training."""

    @property
    @abc.abstractmethod
    @functools.lru_cache
    def unique_dataset_item_ids(self) -> np.array:
        """Cached unique sorted item ids that the model can encounter during training."""

    @property
    @functools.lru_cache()
    def shape(self) -> "tuple[int, int]":
        n_users = len(self.unique_dataset_user_ids)
        n_items = len(self.unique_dataset_item_ids)
        return n_users, n_items

    def dense_to_dataset_user_ids(self, dense_user_ids: np.array) -> np.array:
        return self.unique_dataset_user_ids[dense_user_ids]

    def dataset_to_dense_user_ids(self, dataset_user_ids: np.array) -> np.array:
        dataset_to_dense = pd.Series(
            index=self.unique_dataset_user_ids,
            data=np.arange(len(self.unique_dataset_user_ids)),
        )
        return dataset_to_dense.loc[dataset_user_ids].values

    def dense_to_dataset_item_ids(self, dense_item_ids: np.array) -> np.array:
        return self.unique_dataset_item_ids[dense_item_ids]

    def dataset_to_dense_item_ids(self, dataset_item_ids: np.array) -> np.array:
        dataset_to_model = pd.Series(
            index=self.unique_dataset_item_ids,
            data=np.arange(len(self.unique_dataset_item_ids)),
        )
        return dataset_to_model.loc[dataset_item_ids].values


class MovieLensInterface(RecommendingDatasetInterface, abc.ABC):
    def __init__(self, directory):
        self.directory = directory

    def abs_path(self, filename):
        return os.path.join(os.getcwd(), self.directory, filename)

    @property
    def ratings_columns_to_dtypes(self) -> OrderedDict:
        return OrderedDict(
            {
                "user_id": np.int32,
                "item_id": np.int32,
                "rating": np.float32,
                "timestamp": np.int32,
            }
        )

    @abc.abstractmethod
    @functools.lru_cache()
    def __getitem__(self, filename):
        return self.read(filename=filename)

    def read(self, filename, **kwargs):
        path = os.path.join(self.directory, filename)
        dataframe = pd.read_csv(path, **kwargs)
        return dataframe.squeeze()

    def explicit_feedback_scipy_csr(self, name):
        return self.ratings_dataframe_to_scipy_csr(self[name])

    def ratings_dataframe_to_scipy_csr(self, ratings_dataframe):
        movielens_ratings = ratings_dataframe["rating"].to_numpy()

        dataset_user_ids = ratings_dataframe["user_id"].to_numpy()
        user_ids = self.dataset_to_dense_user_ids(dataset_user_ids)

        dataset_item_ids = ratings_dataframe["item_id"].to_numpy()
        item_ids = self.dataset_to_dense_item_ids(dataset_item_ids)

        data = movielens_ratings
        row_ids = user_ids
        col_ids = item_ids
        explicit_feedback = coo_matrix((data, (row_ids, col_ids)), shape=self.shape)
        return explicit_feedback.tocsr()

    @abc.abstractmethod
    def items_description(self, dense_item_ids: np.array) -> pd.DataFrame:
        """Generates dataframe of items description with an item per row in free form."""

    @staticmethod
    def construct_1d_explicit(
        item_ids: np.array, ratings: np.array, n_items
    ) -> "SparseTensor":
        return torch.sparse_coo_tensor(
            indices=np.stack([np.zeros_like(item_ids), item_ids]),
            values=ratings,
            size=[1, n_items],
        )

    @abc.abstractmethod
    def imdb_ratings_dataframe_to_explicit(
        self, imdb_ratings: pd.DataFrame
    ) -> "SparseTensor":
        """
        Builds explicit feedback matrix from imdb_ratings DataFrame
        returned by the csv_imdb_ratings_to_dataframe function.
        """


class MovieLens100k(MovieLensInterface):
    def __getitem__(self, filename):
        if filename == "u.info":
            return self.read(
                filename=filename,
                names=["quantity", "index"],
                index_col="index",
                sep=" ",
            )
        elif filename == "u.genre":
            return self.read(filename=filename, names=["name", "id"], index_col="id")
        elif filename == "u.occupation":
            return self.read(filename=filename, names=["occupation"])
        elif filename == "u.user":
            return self.read(
                filename=filename,
                names=["user_id", "age", "gender", "occupation", "zip code"],
                index_col="user_id",
            )
        elif filename == "u.item":
            return self.read(
                filename=filename,
                names=[
                    "movie id",
                    "movie title",
                    "release date",
                    "video release date",
                    "IMDb URL",
                    "unknown",
                    "Action",
                    "Adventure",
                    "Animation",
                    "Children's",
                    "Comedy",
                    "Crime",
                    "Documentary",
                    "Drama",
                    "Fantasy",
                    "Film-Noir",
                    "Horror",
                    "Musical",
                    "Mystery",
                    "Romance",
                    "Sci-Fi",
                    "Thriller",
                    "War",
                    "Western",
                ],
                index_col="movie id",
                encoding_errors="backslashreplace",
            )
        if filename in ["u.data"] + [
            i + j
            for i in ["u1", "u2", "u3", "u4", "u5", "ua", "ub"]
            for j in [".base", ".test"]
        ]:
            return self.read(
                filename=filename,
                names=list(self.ratings_columns_to_dtypes),
                sep="\t",
                dtype=self.ratings_columns_to_dtypes,
            )
        raise FileNotFoundError(f"File {filename} not found.")

    def read(self, filename, sep="|", header=None, **kwargs):
        return super().read(filename=filename, sep=sep, header=header, **kwargs)

    @property
    @functools.lru_cache
    def unique_dataset_user_ids(self) -> np.array:
        return np.unique(self["u.user"].index)

    @property
    @functools.lru_cache
    def unique_dataset_item_ids(self):
        return np.unique(self["u.item"].index)

    def items_description(self, dense_item_ids: np.array):
        assert dense_item_ids.ndim == 1
        dataset_item_ids = self.dense_to_dataset_item_ids(dense_item_ids)
        description = (
            self["u.item"]
            .loc[dataset_item_ids]
            .reset_index(names="movielens_100k_dataset_ids")
        )
        description["dense_item_ids"] = dense_item_ids
        return description

    @staticmethod
    def normalize_titles(titles):
        """
        Normalizes unicode string, removes punctuation, article 'the',
        deduplicates consecutive spaces and strips fringing spaces.
        """
        titles = (
            titles.str.lower()
            .str.normalize("NFC")
            .str.replace(f"[{string.punctuation}]", "", regex=True)
            .str.replace("the", "")
            .str.replace(r"\s+", " ", regex=True)
            .str.strip()
        )
        return titles

    def imdb_ratings_dataframe_to_explicit(self, imdb_ratings: pd.DataFrame):
        imdb_ratings["title"] = self.normalize_titles(imdb_ratings["title"])

        movielens_titles = self["u.item"]["movie title"]
        movielens_titles = self.normalize_titles(movielens_titles.str.split("(").str[0])

        movielens_movie_ids = []
        ratings = []

        for imdb_title, rating in zip(
            imdb_ratings["title"], imdb_ratings["your_rating"]
        ):
            for movielens_movie_id, movielens_title in movielens_titles.items():
                if imdb_title == movielens_title:
                    movielens_movie_ids.append(movielens_movie_id)
                    ratings.append(rating)

        ratings = np.array(ratings)
        movielens_movie_ids = np.array(movielens_movie_ids)
        item_ids = self.dataset_to_dense_item_ids(movielens_movie_ids)

        return self.construct_1d_explicit(
            item_ids=item_ids, ratings=ratings, n_items=self.shape[1]
        )


class MovieLens25m(MovieLensInterface):
    @profile()
    def prepare_splits(self, train_time_quantile: float, test_user_fraction: float):
        time_split_ante, time_split_post = dataframe_quantiles_deterministic_split(
            quantiles=[train_time_quantile, 1 - train_time_quantile],
            dataframe=self["ratings"],
            column_to_split_by="timestamp",
        )
        unique_user_ids = time_split_post["user_id"].unique()
        test_user_ids = np.random.choice(
            unique_user_ids,
            size=int(len(unique_user_ids) * test_user_fraction),
            replace=False,
        )
        val_split = time_split_post.query(
            "user_id not in @test_user_ids",
            local_dict=dict(test_user_ids=test_user_ids),
        )
        test_split = time_split_post.query(
            "user_id in @test_user_ids",
            local_dict=dict(test_user_ids=test_user_ids),
        )
        splits = dict(train=time_split_ante, val=val_split, test=test_split)
        for kind, split in splits.items():
            filename = f"ratings_{kind}.csv"
            split.to_csv(path_or_buf=self.abs_path(filename=filename), index=False)

    @functools.lru_cache(maxsize=20)
    def __getitem__(self, filename):
        if not filename.endswith(".csv"):
            filename += ".csv"

        if filename == "ratings.csv":
            return self.read(
                filename=filename,
                names=list(self.ratings_columns_to_dtypes),
                header=0,
                dtype=self.ratings_columns_to_dtypes,
            )
        elif filename == "links.csv":
            nullable_int64 = pd.Int64Dtype()
            return self.read(
                filename=filename,
                dtype={
                    "movieId": np.int32,
                    "imdbId": np.int32,
                    "tmdbId": nullable_int64,
                },
            ).rename({"movieId": "movielensId"}, axis="columns")
        elif filename == "movies.csv":
            return self.read(
                filename=filename,
                index_col="movieId",
                dtype={"movieId": np.int32, "title": str, "genres": str},
            ).rename_axis("item_id")
        elif filename == "tags.csv":
            dtypes = self.ratings_columns_to_dtypes
            dtypes.pop("rating")
            return (
                self.read(filename=filename, dtype={"tag": str})
                .rename({"userId": "user_id", "movieId": "item_id"}, axis="columns")
                .astype(dtypes)
                .dropna()
            )
        else:
            return self.read(filename=filename)

    @property
    @functools.lru_cache
    def unique_dataset_user_ids(self) -> np.array:
        return np.unique(self["ratings"]["user_id"])

    @property
    @functools.lru_cache
    def unique_dataset_item_ids(self):
        return np.unique(self["ratings"]["item_id"])

    @property
    @functools.lru_cache()
    def unique_imdb_ids(self):
        return np.unique(self["links"]["imdbId"])

    def dataset_to_imdb_movie_ids(self, dataset_item_ids: np.array):
        movielens_to_imdb = self["links"].set_index("movielensId")["imdbId"]
        return movielens_to_imdb.loc[dataset_item_ids].values

    def imdb_to_dataset_item_ids(self, imdb_movie_ids: np.array):
        imdb_to_movielens = self["links"].set_index("imdbId")["movielensId"]
        return imdb_to_movielens.loc[imdb_movie_ids].values

    def items_description(self, dense_item_ids: np.array):
        assert dense_item_ids.ndim == 1
        dataset_item_ids = self.dense_to_dataset_item_ids(dense_item_ids)
        description = (
            self["movies"]
            .loc[dataset_item_ids]
            .reset_index(names="movielens_movie_ids")
        )
        description["dense_item_ids"] = dense_item_ids
        description = pd.merge(
            description,
            self["links"],
            how="left",
            left_on="movielens_movie_ids",
            right_on="movielensId",
        )
        return description

    def imdb_ratings_dataframe_to_explicit(self, imdb_ratings: pd.DataFrame):
        imdb_ids = imdb_ratings["imdb_id"].values
        imdb_ids = np.intersect1d(self.unique_imdb_ids, imdb_ids)
        dataset_ids = self.imdb_to_dataset_item_ids(imdb_ids)
        item_ids = self.dataset_to_dense_item_ids(dataset_ids)
        ratings = imdb_ratings.set_index("imdb_id").loc[imdb_ids]["your_rating"].values
        return self.construct_1d_explicit(
            item_ids=item_ids, ratings=ratings, n_items=self.shape[1]
        )


class MovieLensMixin:
    @property
    def movielens(self: "LitRecommenderBase") -> MovieLensInterface:
        return self.build_class(**self.hparams["datamodule"]["movielens"])

    @property
    def class_candidates(self):
        return super().class_candidates + [MovieLens100k, MovieLens25m]

    def common_explicit(self: "LitRecommenderBase", filename):
        if (movielens_inner_name := self.hparams["datamodule"][filename]) is None:
            return
        try:
            return self.movielens.explicit_feedback_scipy_csr(name=movielens_inner_name)
        except FileNotFoundError:
            return

    def train_explicit(self):
        return self.common_explicit(filename="train_explicit_file")

    def val_explicit(self):
        return self.common_explicit(filename="val_explicit_file")

    def test_explicit(self):
        return self.common_explicit(filename="test_explicit_file")


def dataframe_quantiles_deterministic_split(
    dataframe: pd.DataFrame, quantiles: "list[float]", column_to_split_by: str
) -> "list[pd.DataFrame]":
    if sum(quantiles) != 1:
        raise ValueError(f"Quantiles must sum up to one, received {quantiles}.")
    if not pd.api.types.is_numeric_dtype(dataframe[column_to_split_by].dtype):
        raise ValueError(
            f"Splitting column's dtype isn't numeric: "
            f"{column_to_split_by} -> {dataframe[column_to_split_by].dtype} "
        )
    if any(dataframe[column_to_split_by].isna()):
        raise ValueError(f"Splitting column {column_to_split_by} contains nans.")

    quantile_thresholds = np.cumsum(quantiles)
    thresholds = dataframe[column_to_split_by].quantile(
        quantile_thresholds, interpolation="lower"
    )
    thresholds = [-np.inf] + list(thresholds)
    splits = []
    for begin, end in zip(thresholds, thresholds[1:]):
        splits.append(dataframe.query(f"{begin} < {column_to_split_by} <= {end}"))

    assert sum(map(len, splits)) == len(dataframe), "Split lengths don't sum up."
    return splits


def main():
    parser = argparse.ArgumentParser(
        description="Prepare movielens25m artifact and splits."
    )
    parser.add_argument("directory", type=str, help="movielens25m directory")
    parser.add_argument("train_time_quantile", type=float)
    parser.add_argument("test_user_fraction", type=float)
    parser.add_argument(
        "update_with_local_directory", type=bool, nargs="?", default=False
    )
    args = parser.parse_args()

    movielens25m = MovieLens25m(directory=args.directory)
    wandb.init(project="Recommending", job_type="data")
    artifact = prepare_artifact(
        full_artifact_name="dkoshman/Recommending/movielens25m:latest",
        directory=args.directory,
        artifact_type="data",
        update_with_local_directory=args.update_with_local_directory,
    )
    artifact.wait()
    movielens25m.prepare_splits(
        train_time_quantile=args.train_time_quantile,
        test_user_fraction=args.test_user_fraction,
    )
    artifact.metadata["train_time_quantile"] = args.train_time_quantile
    artifact.metadata["test_user_fraction"] = args.test_user_fraction
    artifact.save()


if __name__ == "__main__":
    main()



from .entrypoint import RecommendingConfigDispenser, fit
from .utils import update_from_base_config, wandb_context_manager


def cross_validation_configs(config):
    for i in [1, 2, 3, 4, 5]:
        config["datamodule"].update(
            dict(
                train_explicit_file=f"u{i}.base",
                val_explicit_file=f"u{i}.test",
                test_explicit_file=f"u{i}.test",
            )
        )
        yield config


def main(config):
    config = update_from_base_config(
        config, base_config_file="configs/base_config.yaml"
    )
    if not config["wandb_save"]:
        if "WandbCheckpointCallback" in config.get("callbacks", []):
            del config["callbacks"]["WandbCheckpointCallback"]
    if config["cross_validate"]:
        if "CatBoostMetrics" in config.get("callbacks", []):
            del config["callbacks"]["CatBoostMetrics"]
        with wandb_context_manager(config):
            for config in cross_validation_configs(config):
                fit(config)
    else:
        fit(config)


if __name__ == "__main__":
    dispenser = RecommendingConfigDispenser()
    dispenser.from_cli()
    dispenser.launch(main)



raise NotImplementedError("There's something wrong with this implementation")

from typing import Literal

import einops
import numba
import numpy as np
import scipy
import torch

from tqdm.auto import tqdm

from my_tools.models import WandbLoggerMixin

from ..interface import RecommenderModuleBase, FitExplicitInterfaceMixin


class BayesianPMF(RecommenderModuleBase, FitExplicitInterfaceMixin, WandbLoggerMixin):
    def __init__(
        self,
        n_feature_dimensions=10,
        burn_in_steps=500,
        keeper_steps=100,
        predictive_explicit_precision=1,
        features_hyper_precision_coefficient=1,
        **kwargs,
    ):
        """
        This model approximates not just a single point estimate in parameter space,
        but a whole distribution, therefore optimizing over hyperparameter space.
        It accomplishes this by MCMC sampling to estimate intractable integral
        that represents the probability of observing the train data:
            P(X) = \\int P(X|Z)P(Z)dZ \\approx \\sum_i^n P(X|Z_i) / n
        where Z_i are sampled from some prior distribution, and uses it to maximize P(Z|X):
            P(Z|X) = P(X|Z)P(Z) / P(X)
        In the end though we still need a single estimate point, which
        is usually the mean or the mode of distribution P(Z|X).

        See https://www.cs.toronto.edu/~amnih/papers/bpmf.pdf
        https://en.wikipedia.org/wiki/Markov_chain_Monte_Carlo
        https://en.wikipedia.org/wiki/Normal-Wishart_distribution

        :param n_feature_dimensions: size of latent dimension
        :param burn_in_steps: first samples are usually pretty biased,
        so we skip them in hopes that after that model has substantially converged
        :param keeper_steps: the steps after burn in to run for
        :param predictive_explicit_precision: the shared precision of normal distributions
        (whose means will be saved in step_explicit_normal_distribution_means),
        which comprise the uniformly weighted mixture that is the estimation of
        P(Y|X)=f(Z)P(Z|X), where Y=f(z) are the predicted ratings
        :param features_hyper_precision_coefficient: hyper-hyperparameters precision
        distribution, the higher it is, the more regularized the model will be;
        it corresponds to lambda in the last wiki link
        """

        super().__init__(**kwargs)
        self.n_feature_dimensions = n_feature_dimensions
        self.burn_in_steps = burn_in_steps
        self.keeper_steps = keeper_steps
        self.alpha = predictive_explicit_precision

        # Coefficients for distribution of hyperparameters
        self.beta = features_hyper_precision_coefficient
        self.hyper_mean = self.init_hyper_mean()
        self.degrees_of_freedom = self.init_degrees_of_freedom()
        self.scale_matrix = self.init_scale_matrix()
        assert self.scale_matrix.shape == (
            self.n_feature_dimensions,
            self.n_feature_dimensions,
        )
        self.hyper_variance = np.linalg.inv(self.scale_matrix)

        self.user_features = self.init_user_features()
        self.item_features = self.init_item_features()

        # The output of the model – means of normal distributions
        # with precision predictive_explicit_precision, the mixture
        # of which is the estimated distribution of explicit feedback.
        self.step_explicit_normal_distribution_means = torch.nn.Parameter(
            torch.zeros(keeper_steps, self.n_users, self.n_items), requires_grad=False
        )

        self.implicit = None
        self.explicit = None
        self.alpha_x_implicit_x_explicit = None

    def init_hyper_mean(self):
        return 0

    def init_degrees_of_freedom(self):
        return self.n_feature_dimensions

    def init_scale_matrix(self):
        return np.eye(self.n_feature_dimensions)

    def init_features(self, size):
        return np.random.multivariate_normal(
            mean=np.full(self.n_feature_dimensions, self.hyper_mean),
            cov=self.hyper_variance / self.n_feature_dimensions,
            size=size,
        )

    def init_user_features(self):
        """Override this method if you want to finetune a fitted MF model."""
        return self.init_features(self.n_users)

    def init_item_features(self):
        """Override this method if you want to finetune a fitted MF model."""
        return self.init_features(self.n_items)

    def fit(self, explicit):
        self.explicit = explicit
        self.implicit = explicit > 0
        self.alpha_x_implicit_x_explicit = self.alpha * self.implicit.multiply(explicit)

        for _ in tqdm(range(self.burn_in_steps), "Sampling burn in steps"):
            self.step()

        for step in tqdm(range(self.keeper_steps), "Sampling keeper steps"):
            step_means = torch.from_numpy(self.step())
            self.step_explicit_normal_distribution_means[step] = step_means

    def step(self):
        user_hyperparameters = self.sample_hyperparameters(self.user_features)
        item_hyperparameters = self.sample_hyperparameters(self.item_features)

        user_features = self.sample_features(
            self.item_features,
            self.implicit,
            self.alpha_x_implicit_x_explicit,
            **user_hyperparameters,
        )
        item_features = self.sample_features(
            self.user_features,
            self.implicit.T.tocsr(),
            self.alpha_x_implicit_x_explicit.T.tocsr(),
            **item_hyperparameters,
        )

        self.user_features = user_features
        self.item_features = item_features

        step_explicit_mean = user_features @ item_features.T

        self.log(
            dict(
                mse=((step_explicit_mean - self.explicit.toarray()) ** 2).mean(),
                mse_filtered=(
                    (
                        self.implicit.multiply(step_explicit_mean).toarray()
                        - self.explicit.toarray()
                    )
                    ** 2
                ).mean(),
            )
        )
        return step_explicit_mean

    def sample_hyperparameters(self, features):
        n = features.shape[0]
        features_mean = np.einsum("i d -> d", features) / n
        sample_variance = np.einsum(
            "i d, i k -> d k", features - features_mean, features - features_mean
        )

        beta_estimation = self.beta + n
        degrees_of_freedom_estimation = self.degrees_of_freedom + n
        hyper_mean_estimation = (self.beta * self.hyper_mean + n * features_mean) / (
            self.beta + n
        )
        biased_variance_estimation = np.einsum(
            "d, k -> d k",
            features_mean - self.hyper_mean,
            features_mean - self.hyper_mean,
        )
        scale_matrix_estimation = np.linalg.inv(
            self.hyper_variance
            + sample_variance
            + (self.beta * n) / (self.beta + n) * biased_variance_estimation
        )

        hyper_variance_inverse = scipy.stats.wishart.rvs(
            df=degrees_of_freedom_estimation, scale=scale_matrix_estimation
        )
        hyper_mean = np.random.multivariate_normal(
            hyper_mean_estimation,
            np.linalg.inv(beta_estimation * hyper_variance_inverse),
        )
        return dict(
            hyper_variance_inverse=hyper_variance_inverse, hyper_mean=hyper_mean
        )

    def sample_features(
        self,
        fixed_features,
        implicit,
        alpha_x_implicit_x_ratings,
        hyper_mean,
        hyper_variance_inverse,
    ):
        n = implicit.shape[0]
        means = np.empty((n, self.n_feature_dimensions))
        variances = np.empty((n, self.n_feature_dimensions, self.n_feature_dimensions))
        alpha_x_features_x_features_t = self.alpha * np.einsum(
            "j d, j k -> j d k", fixed_features, fixed_features
        )

        precalculated = (
            hyper_variance_inverse @ hyper_mean
            + alpha_x_implicit_x_ratings @ fixed_features
        )

        self.jit_sample_features(
            precalculated,
            alpha_x_features_x_features_t,
            hyper_variance_inverse,
            implicit.indices,
            implicit.indptr,
            means,
            variances,
        )

        distribution = torch.distributions.MultivariateNormal(
            loc=torch.from_numpy(means), covariance_matrix=torch.from_numpy(variances)
        )
        features = distribution.sample().numpy()
        return features

    @staticmethod
    @numba.njit(parallel=True)
    def jit_sample_features(
        precalculated,
        alpha_x_features_x_features_t,
        hyper_variance_inverse,
        indices,
        indptr,
        means,
        variances,
    ):
        hyper_variance = np.linalg.inv(hyper_variance_inverse)
        for i in numba.prange(len(means)):
            col_indices = indices[indptr[i] : indptr[i + 1]]
            if col_indices.size:
                var = np.linalg.inv(
                    hyper_variance_inverse
                    + alpha_x_features_x_features_t[col_indices].sum(axis=0)
                )
            else:
                var = hyper_variance
            mean = var @ precalculated[i]
            means[i] = mean
            variances[i] = var

    def aggregate_prediction_mean(self, user_ids):
        return einops.reduce(
            self.step_explicit_normal_distribution_means[:, user_ids, :],
            "step user item -> user item",
            "mean",
        )

    def aggregate_prediction_mode_approximation(self, user_ids):
        # Caution: very memory hungry.
        means = self.step_explicit_normal_distribution_means[:, user_ids, :]

        normal = torch.distributions.Normal(loc=means, scale=self.alpha**-0.5)
        pdfs_at_means = normal.log_prob(
            einops.repeat(means, f"step user item -> step {means.shape[0]} user item")
        )
        step_ids_with_highest_mean_pdf = torch.logsumexp(pdfs_at_means, dim=0).argmax(0)
        mode_means = torch.gather(means, 0, step_ids_with_highest_mean_pdf[None])[0]
        return mode_means

    def forward(self, user_ids, aggregation_method: Literal["mean", "mode"] = "mean"):
        match aggregation_method:
            case "mean":
                ratings = self.aggregate_prediction_mean(user_ids)
            case "mode":
                ratings = self.aggregate_prediction_mode_approximation(user_ids)
            case _:
                raise ValueError(f"Unknown aggregation method {aggregation_method}")
        return ratings



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['my_recommending', 'my_recommending.models', 'my_recommending.movielens']

package_data = \
{'': ['*']}

install_requires = \
['aenum>=3.1.11,<4.0.0',
 'catboost>=1.1,<2.0',
 'dkoshman-my-tools>=0.1.4,<0.2.0',
 'einops>=0.4.1,<0.5.0',
 'gradio>=3.7,<4.0',
 'implicit>=0.6.1,<0.7.0',
 'line-profiler>=4.0.2,<5.0.0',
 'matplotlib>=3.5.0,<3.6.0',
 'numba>=0.56.2,<0.57.0',
 'numpy>=1.22,<2.0',
 'pandas>=1.4.0,<2.0.0',
 'protobuf==3.19.4',
 'pytest>=7.1.3,<8.0.0',
 'pytorch-lightning>=1.6,<1.7',
 'scikit-learn>=1.2.0,<2.0.0',
 'scipy>=1.7,<2.0',
 'shap>=0.41.0,<0.42.0',
 'torch>=1.12.1,<2.0.0',
 'trash-cli>=0.22,<0.23',
 'wandb>=0.13.3,<0.14.0']

setup_kwargs = {
    'name': 'my-recommending',
    'version': '0.1.17',
    'description': '',
    'long_description': 'Recommending module with svd, matrix factorization and catboost recommendation models.\n',
    'author': 'DimaKoshman',
    'author_email': 'koshmandk@yandex.ru',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)
